@@ -1,6 +1,6 @@
 #
 # ChrootUnshare.pm: chroot library for sbuild
-# Copyright © 2018      Johannes Schauer <josch@debian.org>
+# Copyright © 2018      Johannes Schauer Marin Rodrigues <josch@debian.org>
 #
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -71,7 +71,7 @@ sub begin_session {
 	}
 	$tarball = abs_path($chroot);
     } else {
-	my $xdg_cache_home = $ENV{'HOME'} . "/.cache/sbuild";
+	my $xdg_cache_home = $self->get_conf('HOME') . "/.cache/sbuild";
 	if (defined($ENV{'XDG_CACHE_HOME'})) {
 	    $xdg_cache_home = $ENV{'XDG_CACHE_HOME'} . '/sbuild';
 	}
@@ -105,9 +105,16 @@ sub begin_session {
     my @idmap = read_subuid_subgid;
 
     # sanity check
-    if (scalar(@idmap) != 2 || $idmap[0][0] ne 'u' || $idmap[1][0] ne 'g') {
-	printf STDERR "invalid idmap\n";
-	return 0;
+    if (   scalar(@idmap) != 2
+        || $idmap[0][0] ne 'u'
+        || $idmap[1][0] ne 'g'
+        || length $idmap[0][1] == 0
+        || length $idmap[0][2] == 0
+        || length $idmap[1][1] == 0
+        || length $idmap[1][2] == 0)
+    {
+        printf STDERR "invalid idmap\n";
+        return 0;
     }
 
     $self->set('Uid Gid Map', \@idmap);
@@ -163,6 +170,7 @@ sub begin_session {
 	'--exclude=./dev/random',
 	'--exclude=./dev/full',
 	'--exclude=./dev/null',
+	'--exclude=./dev/console',
 	'--exclude=./dev/zero',
 	'--exclude=./dev/tty',
 	'--exclude=./dev/ptmx',
@@ -190,13 +198,6 @@ sub begin_session {
 	return 0;
     }
 
-    # some initial setup
-    for my $user ($self->get_conf('USERNAME'), $self->get_conf('BUILD_USER')) {
-	system('env', 'PATH=/usr/sbin:/usr/bin:/sbin:/bin', @unshare_cmd,
-	    '/usr/sbin/chroot', $rootdir, 'sh', '-c',
-	    "id -u \"$user\">/dev/null 2>&1 || adduser --system --quiet --ingroup sbuild --no-create-home --home /nonexistent --disabled-login --disabled-password \"$user\"");
-    }
-
     $self->set('Session ID', $rootdir);
 
     $self->set('Location', '/sbuild-unshare-dummy-location');
@@ -270,35 +271,77 @@ sub _get_exec_argv {
     my $dir = shift;
     my $user = shift;
     my $disable_network = shift // 0;
+    my $build_inside_init = shift // 0;
 
-    my $network_setup = 'cat /etc/resolv.conf > "$rootdir/etc/resolv.conf";';
+    # On systems with libnss-resolve installed there is no need for a
+    # /etc/resolv.conf. This works around this by adding 127.0.0.53 (default
+    # for systemd-resolved) in that case.
+    my $network_setup = '[ -f /etc/resolv.conf ] && cat /etc/resolv.conf > "$rootdir/etc/resolv.conf" || echo "nameserver 127.0.0.53" > "$rootdir/etc/resolv.conf";';
     my $unshare = CLONE_NEWNS | CLONE_NEWPID | CLONE_NEWUTS | CLONE_NEWIPC;
+    my $init = "";
     if ($disable_network) {
 	$unshare |= CLONE_NEWNET;
 	$network_setup = 'ip link set lo up;> "$rootdir/etc/resolv.conf";';
     }
 
+    if ($build_inside_init) {
+	$init = "/usr/bin/dumb-init";
+    }
+
+    # Detect whether linux32 personality might be needed
+    my %personalities = (
+	'armel:arm64'     => 1,
+	'armhf:arm64'     => 1,
+	'i386:amd64'      => 1,
+	'mipsel:mips64el' => 1,
+	'powerpc:ppc64'   => 1,
+	's390:s390x'      => 1,
+	'sparc:sparc64'   => 1,
+    );
+    my $linux32 = exists $personalities{($self->get_conf('BUILD_ARCH') . ':' . $self->get_conf('ARCH'))};
+
+    my @bind_mounts = ();
+    for my $entry (@{$self->get_conf('UNSHARE_BIND_MOUNTS')}) {
+	push @bind_mounts, $entry->{directory}, $entry->{mountpoint};
+    }
+
     return (
-	'env', 'PATH=/usr/sbin:/usr/bin:/sbin:/bin',
-	get_unshare_cmd({UNSHARE_FLAGS => $unshare, FORK => 1, IDMAP => $self->get('Uid Gid Map')}), 'sh', '-c', "
+	'env', 'PATH=' . $self->get_conf('PATH'), "USER=$user", "LOGNAME=$user",
+	get_unshare_cmd({UNSHARE_FLAGS => $unshare, FORK => 1, IDMAP => $self->get('Uid Gid Map'), LINUX32 => $linux32}), 'sh', '-c', "
 	rootdir=\"\$1\"; shift;
 	user=\"\$1\"; shift;
 	dir=\"\$1\"; shift;
+	while [ \$# -gt 0 ]; do
+	    if [ \"\$1\" = \"--\" ]; then shift; break; fi;
+	    mkdir -p \"\$rootdir\$2\";
+	    mount -o rbind \"\$1\" \"\$rootdir\$2\";
+	    shift; shift;
+	done;
 	hostname sbuild;
+	echo \"127.0.0.1 localhost\\n127.0.1.1 sbuild\\n::1 localhost ip6-localhost ip6-loopback\" > \"\$rootdir/etc/hosts\";
 	$network_setup
 	mkdir -p \"\$rootdir/dev\";
-	for f in null zero full random urandom tty; do
+	for f in null zero full random urandom tty console; do
 	    touch \"\$rootdir/dev/\$f\";
 	    chmod -rwx \"\$rootdir/dev/\$f\";
 	    mount -o bind \"/dev/\$f\" \"\$rootdir/dev/\$f\";
 	done;
+	ln -sfT /proc/self/fd \"\$rootdir/dev/fd\";
+	ln -sfT /proc/self/fd/0 \"\$rootdir/dev/stdin\";
+	ln -sfT /proc/self/fd/1 \"\$rootdir/dev/stdout\";
+	ln -sfT /proc/self/fd/2 \"\$rootdir/dev/stderr\";
+	mkdir -p \"\$rootdir/dev/pts\";
+	mount -o noexec,nosuid,gid=5,mode=620,ptmxmode=666 -t devpts none \"\$rootdir/dev/pts\";
+	ln -sfT /dev/pts/ptmx \"\$rootdir/dev/ptmx\";
+	mkdir -p \"\$rootdir/dev/shm\";
+	mount -t tmpfs tmpfs \"\$rootdir/dev/shm\";
 	mkdir -p \"\$rootdir/sys\";
 	mount -o rbind /sys \"\$rootdir/sys\";
+	mount -t tmpfs tmpfs \"\$rootdir/sys/kernel\" -o mode=0000,size=4k,ro;
 	mkdir -p \"\$rootdir/proc\";
 	mount -t proc proc \"\$rootdir/proc\";
-	/usr/sbin/chroot \"\$rootdir\" sh -c \"id -u \\\"\$user\\\">/dev/null 2>&1 || adduser --system --quiet --ingroup sbuild --no-create-home --home /nonexistent --disabled-login --disabled-password \\\"\$user\\\"\";
-	exec /usr/sbin/chroot \"\$rootdir\" /sbin/runuser -u \"\$user\" -- sh -c \"cd \\\"\\\$1\\\" && shift && \\\"\\\$@\\\"\" -- \"\$dir\" \"\$@\";
-	", '--', $self->get('Session ID'), $user, $dir
+	exec /usr/sbin/chroot \"\$rootdir\" $init /sbin/runuser -p -u \"\$user\" -- sh -c \"cd \\\"\\\$1\\\" && shift && \\\"\\\$@\\\"\" -- \"\$dir\" \"\$@\";
+	", '--', $self->get('Session ID'), $user, $dir, @bind_mounts, '--'
     );
 }
 
@@ -334,11 +377,16 @@ sub get_command_internal {
     }
 
     my $disable_network = 0;
-    if (defined($options->{'DISABLE_NETWORK'}) && $options->{'DISABLE_NETWORK'}) {
+    if (defined($options->{'ENABLE_NETWORK'}) && $options->{'ENABLE_NETWORK'} == 0) {
 	$disable_network = 1;
     }
 
-    my @cmdline = $self->_get_exec_argv($dir, $user, $disable_network);
+    my $build_inside_init = 0;
+    if (defined($options->{'BUILD_INSIDE_INIT'}) && $options->{'BUILD_INSIDE_INIT'}) {
+	$build_inside_init = 1;
+    }
+
+    my @cmdline = $self->_get_exec_argv($dir, $user, $disable_network, $build_inside_init);
     if (ref $command) {
 	push @cmdline, @$command;
     } else {
@@ -352,4 +400,23 @@ sub get_command_internal {
     $options->{'DIR'} = $dir;
 }
 
+# create users from outside the chroot so we don't need user/groupadd inside.
+sub useradd {
+    my $self = shift;
+    my @args = @_;
+    my $rootdir = $self->get('Session ID');
+    my @idmap = read_subuid_subgid;
+    my @unshare_cmd = get_unshare_cmd({IDMAP => \@idmap});
+    return system(@unshare_cmd, "/usr/sbin/useradd", "--no-log-init", "--prefix", $rootdir, @args);
+}
+
+sub groupadd {
+    my $self = shift;
+    my @args = @_;
+    my $rootdir = $self->get('Session ID');
+    my @idmap = read_subuid_subgid;
+    my @unshare_cmd = get_unshare_cmd({IDMAP => \@idmap});
+    return system(@unshare_cmd, "/usr/sbin/groupadd", "--prefix", $rootdir, @args);
+}
+
 1;
